package owg.lgm;

import owg.lgm.util.AutoDestructibleImpl;
import owg.lgm.util.FILE;
import owg.lgm.util.FramebufferManager;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.IImage.Type;
import owg.lgm.gl.MagFilter;
import owg.lgm.gl.MinFilter;
import owg.lgm.gl.Tex2D;
import owg.lgm.gl.TexFormat;
import owg.lgm.gl.TexSampler;
import owg.lgm.gl.Wrap;
import owg.lgm.plaf.PNGIO;
import owg.util.autoname.AutoName;
import owg.util.fileextensions.FileExtensions;
import owg.util.math.Clamp;
import owg.util.math.MinMax;
import owg.util.math.Modulo;

public class Sprite extends AutoDestructibleImpl {
	public static final HashMap<String, Sprite> registry = new HashMap<>();

	public static void load(GL2ES1 gl, FILE current) throws IOException
	{
		String ext = FileExtensions.getFileExtension(current.getPath());
		if(!"png".equals(ext))
			return;
		//System.out.println("Loading Sprite: "+current);
		IImage image = PNGIO.load(current);
		int nk = image.name.length();
		int k = image.name.lastIndexOf('.', nk-2);
		
		int x = 0, y = 0, l = 0, t = 0, r = image.width-1, n = 1;
		Integer b = null;
		boolean q = false;
		
		while(k != -1)
		{
			try {
				char c = image.name.charAt(k+1);
				if(c == 'x')
					x = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 'y')
					y = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 'l')
					l = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 't')
					t = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 'r')
					r = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 'b')
					b = Integer.parseInt(image.name.substring(k+2, nk));
				else if(c == 'q' && nk == k+2)
					q = true;
				else 
					n = Integer.parseInt(image.name.substring(k+1, nk));
			} catch(NumberFormatException e) {
				//It's fine
			}
			nk = k;
			k = image.name.lastIndexOf('.', k-1);
		}
		if(b == null)
			b = (image.height/n)-1;
		Sprite spr = new Sprite(gl, image.name.substring(0, nk), image, n, false);
		spr.set_offset(x, y);
		spr.set_bbox(l, t, r, b);
		if(q)
			spr.precise = false;
		if(registry.put(spr.name, spr) != null)
			System.err.println("Sprite name collision: "+spr.name);
	}
	
	public static Sprite add(GL2ES1 gl, FILE file, int number, boolean precise, boolean smooth, int xorig, int yorig)
	{
		try {
			IImage image = PNGIO.load(file);
			Sprite spr = new Sprite(gl, null, image, number, smooth);
			if(!precise)
				spr.precise = false;
			spr.set_offset(xorig, yorig);
			return spr;
		} catch(IOException err) {
			err.printStackTrace();
			return null;
		}
	}
	
	public static Sprite get(String name)
	{
		Sprite r = registry.get(name);
		if(r == null)
			throw new ResourceException("No such sprite: "+name);
		return r;
	}
	
	public final String name;
	public int width, height, number;
	protected TexFormat format;
	protected int cc;
	
	public boolean precise;
	protected TexSampler sampler;
	
	protected int xOffset, yOffset;
	protected int left, right, top, bottom;
	protected int x0, x1, y0, y1;
	
	public Tex2D[] textures;
	
	protected Sprite(GL2ES1 gl, String name, IImage image, int number, boolean smooth)
	{
		this.name = name == null ? AutoName.create(getClass()) : name;
		sampler = smooth ? TexSampler.LINEAR_CLAMP : TexSampler.NEAREST_CLAMP;
		setImage(gl, image, number);
	}
	
	protected void setImage(GL2ES1 gl, IImage image, int number)
	{
		this.width = image.width;
		this.height = image.height/number;
		this.number = number;
		
		precise = image.type!=Type.RGB && image.type!=Type.GRAY;
		format = image.type.texFormat;
		cc = image.type.bytesPerComponent;
		
		x1 = width-1;
		y1 = height-1;
		right = x1;
		bottom = y1;

		textures = new Tex2D[number];
		for(int i = 0; i<number; i++) {
			ByteBuffer texData = ByteBuffer.allocateDirect(width*height*cc).order(ByteOrder.nativeOrder());
			texData.put(image.data, i*width*height*cc, width*height*cc);
			texData.flip();
			textures[i] = new Tex2D(gl, format, texData,
					width, height, false, sampler);
			autoDestroy(textures[i]);
		}
	}
	
	@Override
	public void dispose()
	{
		if(name != null)
			registry.remove(name);
		super.dispose();
	}
	
	public boolean replace(GL2ES1 gl, FILE file, int number, boolean precise, boolean smooth, int xorig, int yorig)
	{
		try {
			IImage image = PNGIO.load(file);
			
			for(int i = 0; i<textures.length; i++) {
				autoDestructibles.remove(textures[i]);
				textures[i].dispose();
			}
			
			setImage(gl, image, number);
			
			if(!precise)
				this.precise = false;
			set_offset(xorig, yorig);
			return true;
		} catch(IOException err) {
			err.printStackTrace();
			return false;
		}
	}

	public void set_offset(int xoff, int yoff)
	{
		this.xOffset = xoff;
		this.yOffset = yoff;
		bboxUpdate();
	}
	public int get_xoffset()
	{
		return xOffset;
	}
	public int get_yoffset()
	{
		return yOffset;
	}
	
	public void set_bbox(int left, int top, int right, int bottom)
	{
		this.left = left;
		this.top = top;
		this.right = right;
		this.bottom = bottom;
		bboxUpdate();
	}
	public int get_bbox_left()
	{
		return left;
	}
	public int get_bbox_right()
	{
		return right;
	}
	public int get_bbox_top()
	{
		return top;
	}
	public int get_bbox_bottom()
	{
		return bottom;
	}
	
	public int getX0()
	{
		return x0;
	}
	public int getX1()
	{
		return x1;
	}
	
	public int getY0()
	{
		return y0;
	}
	public int getY1()
	{
		return y1;
	}

	protected void bboxUpdate()
	{
		x0 = left - xOffset;
		x1 = right+1 - xOffset;
		y0 = top - yOffset;
		y1 = bottom+1 - yOffset;
	}

	public boolean touches(ObjectTransform tf, double xoff, double yoff, double index, Sprite other, ObjectTransform otherTf, double otherIndex)
	{
		//Offset in world space
		double dx = otherTf.x-tf.x-xoff;
		double dy = otherTf.y-tf.y-yoff;
		
		//Offset local to tf
		double ttx = (dx*tf.c - dy*tf.s); // offset dot dir
		double tty = (dx*tf.s + dy*tf.c); //-offset wedge dir
		//Other's dir local to tf
		double tpx = otherTf.c*tf.c + otherTf.s*tf.s; // other.dir dot dir
		double tpy = otherTf.c*tf.s - otherTf.s*tf.c; //-other.dir wedge dir
		
		//Other's bounding box local to tf
		double tlx00 = (ttx + tpx*otherTf.lx0 - tpy * otherTf.ly0)/tf.xScale;
		double tly00 = (tty + tpy*otherTf.lx0 + tpx * otherTf.ly0)/tf.yScale;

		double tlx10 = (ttx + tpx*otherTf.lx1 - tpy * otherTf.ly0)/tf.xScale;
		double tly10 = (tty + tpy*otherTf.lx1 + tpx * otherTf.ly0)/tf.yScale;

		double tlx01 = (ttx + tpx*otherTf.lx0 - tpy * otherTf.ly1)/tf.xScale;
		double tly01 = (tty + tpy*otherTf.lx0 + tpx * otherTf.ly1)/tf.yScale;

		double tlx11 = (ttx + tpx*otherTf.lx1 - tpy * otherTf.ly1)/tf.xScale;
		double tly11 = (tty + tpy*otherTf.lx1 + tpx * otherTf.ly1)/tf.yScale;
		 
		final double bias = 0.5;
		//Try to find a separating axis along tf's edges
		if(MinMax.min(tlx00, tlx10, tlx01, tlx11) > getX1()-bias)
			return false;
		if(MinMax.max(tlx00, tlx10, tlx01, tlx11) < getX0()+bias)
			return false;
		if(MinMax.min(tly00, tly10, tly01, tly11) > getY1()-bias)
			return false;
		if(MinMax.max(tly00, tly10, tly01, tly11) < getY0()+bias)
			return false;

		//Offset local to other
		double otx = -dx*otherTf.c + dy*otherTf.s; // offset dot other.dir
		double oty = -dx*otherTf.s - dy*otherTf.c; //-offset wedge other.dir
		//Tf's dir local to other
		double opx = tpx; // dir dot otherdir
		double opy = -tpy; //-dir wedge other.dir
		
		//Tf's bounding box local to other
		double olx00 = (otx + opx*tf.lx0 - opy * tf.ly0)/otherTf.xScale;
		double oly00 = (oty + opy*tf.lx0 + opx * tf.ly0)/otherTf.yScale;

		double olx10 = (otx + opx*tf.lx1 - opy * tf.ly0)/otherTf.xScale;
		double oly10 = (oty + opy*tf.lx1 + opx * tf.ly0)/otherTf.yScale;

		double olx01 = (otx + opx*tf.lx0 - opy * tf.ly1)/otherTf.xScale;
		double oly01 = (oty + opy*tf.lx0 + opx * tf.ly1)/otherTf.yScale;

		double olx11 = (otx + opx*tf.lx1 - opy * tf.ly1)/otherTf.xScale;
		double oly11 = (oty + opy*tf.lx1 + opx * tf.ly1)/otherTf.yScale;
		
		//Try to find a separating axis along other's edges
		if(MinMax.min(olx00, olx10, olx01, olx11) > other.getX1()-bias)
			return false;
		if(MinMax.max(olx00, olx10, olx01, olx11) < other.getX0()+bias)
			return false;
		if(MinMax.min(oly00, oly10, oly01, oly11) > other.getY1()-bias)
			return false;
		if(MinMax.max(oly00, oly10, oly01, oly11) < other.getY0()+bias)
			return false;

		//There is no separating axis
		if(!precise && !other.precise)
			return true;
		else
		{
			//double minX = Math.floor(Math.max(tf.x0+yoff, otherTf.x0));
			double minY = Math.floor(Math.max(tf.y0+yoff, otherTf.y0));
			//double maxX = Math.ceil(Math.min(tf.x1+xoff, otherTf.x1));
			double maxY = Math.ceil(Math.min(tf.y1+yoff, otherTf.y1));
		
			int ti = Modulo.mod((int)index, number);
			int oi = Modulo.mod((int)otherIndex, other.number);
			
			for(double y = minY; y <= maxY; y++)
			{
				double x0, y0, y1, s;
				
				double tminX = Double.POSITIVE_INFINITY;
				double tmaxX = Double.NEGATIVE_INFINITY;
				if(tf.y10 > tf.y00) { 
					x0 = tf.x00+xoff; y0 = tf.y00+yoff; y1 = tf.y10+yoff; s = tf.s0010; 
				} else { 
					x0 = tf.x10+xoff; y0 = tf.y10+yoff; y1 = tf.y00+yoff; s = tf.s0010; 
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					tminX = Math.min(tminX, x);
					tmaxX = Math.max(tmaxX, x);
				}
					
				if(tf.y01 > tf.y00) {
					x0 = tf.x00+xoff; y0 = tf.y00+yoff; y1 = tf.y01+yoff; s = tf.s0001;
				} else {
					x0 = tf.x01+xoff; y0 = tf.y01+yoff; y1 = tf.y00+yoff; s = tf.s0001;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					tminX = Math.min(tminX, x);
					tmaxX = Math.max(tmaxX, x);
				}
				if(tf.y11 > tf.y01) {
					x0 = tf.x01+xoff; y0 = tf.y01+yoff; y1 = tf.y11+yoff; s = tf.s0111;
				} else {
					x0 = tf.x11+xoff; y0 = tf.y11+yoff; y1 = tf.y01+yoff; s = tf.s0111;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					tminX = Math.min(tminX, x);
					tmaxX = Math.max(tmaxX, x);
				}
				if(tf.y11 > tf.y10) {
					x0 = tf.x10+xoff; y0 = tf.y10+yoff; y1 = tf.y11+yoff; s = tf.s1011;
				} else {
					x0 = tf.x11+xoff; y0 = tf.y11+yoff; y1 = tf.y10+yoff; s = tf.s1011;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					tminX = Math.min(tminX, x);
					tmaxX = Math.max(tmaxX, x);
				}

				double ominX = Double.POSITIVE_INFINITY;
				double omaxX = Double.NEGATIVE_INFINITY;
				if(otherTf.y10 >otherTf.y00) { 
					x0 = otherTf.x00; y0 = otherTf.y00; y1 = otherTf.y10; s = otherTf.s0010; 
				} else { 
					x0 = otherTf.x10; y0 = otherTf.y10; y1 = otherTf.y00; s = otherTf.s0010; 
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					ominX = Math.min(ominX, x);
					omaxX = Math.max(omaxX, x);
				}
				if(otherTf.y01 > otherTf.y00) {
					x0 = otherTf.x00; y0 = otherTf.y00; y1 = otherTf.y01; s = otherTf.s0001;
				} else {
					x0 = otherTf.x01; y0 = otherTf.y01; y1 = otherTf.y00; s = otherTf.s0001;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					ominX = Math.min(ominX, x);
					omaxX = Math.max(omaxX, x);
				}
				if(otherTf.y11 > otherTf.y01) {
					x0 = otherTf.x01; y0 = otherTf.y01; y1 = otherTf.y11; s = otherTf.s0111;
				} else {
					x0 = otherTf.x11; y0 = otherTf.y11; y1 = otherTf.y01; s = otherTf.s0111;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					ominX = Math.min(ominX, x);
					omaxX = Math.max(omaxX, x);
				}
				if(otherTf.y11 > otherTf.y10) {
					x0 = otherTf.x10; y0 = otherTf.y10; y1 = otherTf.y11; s = otherTf.s1011;
				} else {
					x0 = otherTf.x11; y0 = otherTf.y11; y1 = otherTf.y10; s = otherTf.s1011;
				}
				if(y >= y0 && y <= y1) {
					double x = x0 + (y-y0)*s;
					ominX = Math.min(ominX, x);
					omaxX = Math.max(omaxX, x);
				}
				
				double minX = Math.max(tminX, ominX);
				double maxX = Math.min(tmaxX, omaxX);
				
				for(double x = minX; x < maxX; x++)
				{
					if(getBitUnsafe(tf, ti, x-xoff, y-yoff) && other.getBitUnsafe(otherTf, oi, x, y))
						return true;
				}
			}
			return false;
		}
	}
	protected boolean getBitUnsafe(ObjectTransform t, int imageIndex, double wx, double wy) 
	{
		if(!precise)
			return true;
		
		double dx = wx-t.x;
		double dy = wy-t.y;
		
		double pixelXox, pixelYoy;
		
		if(t.angle == 0)
		{
			pixelXox = dx;
			pixelYoy = dy;
		}
		else
		{
			//dot(d, t.dir)
			pixelXox = dx*t.c - dy*t.s;
			//-wedge(d, t.dir)
			pixelYoy = dx*t.s + dy*t.c;
		}
		int pixelX = Clamp.clampH((int) Math.floor(pixelXox/t.xScale + xOffset), 0, width);
		int pixelY = Clamp.clampH((int) Math.floor(pixelYoy/t.yScale + yOffset), 0, height);
		Tex2D sub = textures[imageIndex];
		int index = ( (pixelY)*sub.dataWidth + pixelX + 1 ) * cc - 1; 
		return sub.data.get(index) != 0;
	}
	
	public boolean getBit(ObjectTransform t, int imageIndex, double wx, double wy) 
	{
		double dx = wx-t.x;
		double dy = wy-t.y;
		
		double pixelXox, pixelYoy;
		
		
		if(t.angle == 0)
		{
			pixelXox = dx;
			pixelYoy = dy;
		}
		else
		{
			//dot(d, t.dir)
			pixelXox = dx*t.c - dy*t.s;
			//-wedge(d, t.dir)
			pixelYoy = dx*t.s + dy*t.c;
		}
		int pixelX = (int) Math.floor(pixelXox/t.xScale + xOffset);
		int pixelY = (int) Math.floor(pixelYoy/t.yScale + yOffset);
		
		if (pixelX >= left && pixelX <= right && pixelY >= top && pixelY <= bottom)
		{
			//Special note: precise is only true for data that includes alpha
			if(!precise)
				return true;
			Tex2D sub = textures[imageIndex];
			int index = ( (pixelY)*sub.dataWidth + pixelX + 1 ) * cc - 1; 
			return sub.data.get(index) != 0;
		}
		return false;
	}

	public static Sprite create_from_screen(FramebufferManager fbm, int x0, int y0, int w, int h, boolean precise, boolean smooth, int xorig, int yorig)
	{
		IImage image = fbm.screenshot(x0, y0, w, h);
		Sprite r = new Sprite(fbm.getGL(), null, image, 1, smooth);
		if(!precise)
			r.precise = false;
		r.set_offset(xorig, yorig);
		return r;
	}
	
	@Override
	public String toString()
	{
		return name;
	}
	
	public void setSmooth(boolean smooth)
	{
		sampler = new TexSampler(
				smooth ? MinFilter.LINEAR : MinFilter.NEAREST, 
				smooth ? MagFilter.LINEAR : MagFilter.NEAREST, 
				sampler.wrapS, 
				sampler.wrapT
		);
		updateSampler();
	}

	public void setTextureRepeat(boolean repeat)
	{
		sampler = new TexSampler(
				sampler.minFilter,
				sampler.magFilter,
				repeat ? Wrap.REPEAT : Wrap.CLAMP_TO_EDGE, 
				repeat ? Wrap.REPEAT : Wrap.CLAMP_TO_EDGE
		);
		updateSampler();
	}
	
	public void setSampler(TexSampler sampler)
	{
		this.sampler = sampler;
		updateSampler();
	}

	protected void updateSampler()
	{
		for(int i = 0; i<textures.length; i++)
			textures[i].setSampler(sampler);
	}
}
