package owg.lgm.room;

import static owg.lgm.GMLib.*;
import static owg.lgm.plaf.Engine.EN;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import com.jogamp.opengl.GL2ES1;
import owg.lgm.Activity;
import owg.lgm.Background;
import owg.lgm.Collider;
import owg.lgm.Entity;
import owg.lgm.IImage;
import owg.lgm.ObjectTransform;
import owg.lgm.Sprite;
import owg.lgm.gl.BlendMode;
import owg.lgm.gl.Tex2D;
import owg.lgm.gl.TexSampler;
import owg.lgm.in.InputState;
import owg.lgm.in.KeyListener;
import owg.lgm.in.MouseIntersector;
import owg.lgm.in.MouseListener;
import owg.lgm.room.RoomData.BackgroundInfo;
import owg.lgm.room.RoomData.InstanceInfo;
import owg.lgm.room.RoomData.RoomInfo;
import owg.lgm.room.RoomData.TileInfo;
import owg.lgm.room.RoomData.ViewInfo;
import owg.lgm.util.AutoDestructibleImpl;
import owg.lgm.util.Destructible;
import owg.lgm.util.Reflect;
import owg.lgm.util.SaveGame;
import owg.util.catalog.Catalog;
import owg.util.iorderset.IOrderHashSet;
import owg.util.iorderset.IOrderSet;
import owg.util.math.Clamp;
import owg.util.math.Modulo;

public class Room extends AutoDestructibleImpl implements Activity
{
	public static Room CURRENT;
	protected static IOrderHashSet<Entity> persistentInstances = new IOrderHashSet<>();
	protected static HashMap<String, Room> persistentRooms = new HashMap<>();

	public static void clearPersistence()
	{
		for(Room r : persistentRooms.values())
			r.dispose();
		persistentRooms.clear();
		for(Entity e : persistentInstances)
			e.dispose();
		persistentInstances.clear();
	}
	
	public static void save(OutputStream out) throws IOException
	{
		try (	BufferedOutputStream bos = new BufferedOutputStream(out);
				DataOutputStream dos = new DataOutputStream(bos)) {
			SaveGame game = new SaveGame(true);
			game.write(dos, EN.gameConfig.global);
			game.write(dos, persistentInstances);
			game.write(dos, persistentRooms);
			game.write(dos, CURRENT);
		}
	}
	
	@SuppressWarnings("unchecked")
	public static void load(InputStream in) throws IOException
	{
		try (	BufferedInputStream bin = new BufferedInputStream(in);
				DataInputStream dis = new DataInputStream(bin)) {
			SaveGame game = new SaveGame(false);
			Entity.REGISTER = false;
			Object global = game.read(dis, null);
			IOrderHashSet<Entity> pInstances = (IOrderHashSet<Entity>) game.read(dis, null);
			HashMap<String, Room> pRooms = (HashMap<String, Room>) game.read(dis, null);
			Room cRoom = (Room) game.read(dis, null);
			
			clearPersistence();
			CURRENT.dispose();
			
			Reflect.copy(EN.gameConfig.global, global);
			persistentInstances = pInstances;
			persistentRooms = pRooms;
			CURRENT = cRoom;
			game.restoreAll(EN.gl);
			throw new ActivityChangeException(ActivityChangeType.PAUSE_RESUME, CURRENT);
		} finally {
			Entity.REGISTER = true;
		}
	}

	public static TransitionFunc transition = TransitionKind.NONE;
	public static int transitionSteps = 0;

	public static Room of(RoomData data)
	{
		Room r = persistentRooms.get(data.info.name);
		if(r == null)
			r = new Room(data);
		return r;
	}
	
	public static class BackgroundData extends BackgroundInfo
	{
		protected Background background;
		@Override
		public void setBackgroundName(String backgroundName) {
			super.setBackgroundName(backgroundName);
			background = backgroundName.isEmpty()?null:Background.get(backgroundName);
		}
		public Background getBackground() {
			return background;
		}
		public void setBackground(Background background) {
			this.background = background;
			this.backgroundName = background.name;
		}
	}
	
	public static class ViewData extends ViewInfo
	{
		protected Class<? extends Entity> objectType = null;
		@Override
		public void setObjectName(String objectName) {
			objectType = objectName.isEmpty()?null:object(objectName);
		}
		public Entity getObject() {
			if(objectType != null)
				return instance(objectType);
			return null;
		}
	}

	public int index;
	public RoomInfo info;
	public boolean hasCreationCode;
	public BackgroundData[] backgrounds;
	public ViewData[] views;
	protected InstanceInfo[] instanceInfo;
	protected TileInfo[] tileInfo;
	
	protected boolean initd = false;

	public Catalog<Destructible> active;
	public Catalog<Destructible> inactive;
	protected final ArrayList<Entity> instances = new ArrayList<>(1024);
	public TileManager tiles;
	protected transient final ArrayList<Renderable> sortedInstances = new ArrayList<>(1024);
	
	protected int iterator;
	protected int iteratorMax;
	protected boolean render = false;
	protected int viewIndex = 0;

	protected Transition currentTransition;
	
	protected Room()
	{
		//
	}
	
	public Room(RoomData data)
	{
		active = new Catalog<>();
		inactive = new Catalog<>();
		tiles = new TileManager();
		autoDestroy(tiles);
		
		index = data.index;
		info = Reflect.copy(new RoomInfo(), data.info);
		hasCreationCode = !data.creationCode.isEmpty();
		backgrounds = Reflect.arrayCopy(new BackgroundData[data.backgrounds.length], data.backgrounds);
		views = Reflect.arrayCopy(new ViewData[data.views.length], data.views);
		instanceInfo = data.instances;
		tileInfo = data.tiles;
	}

	@Override
	public void start(GL2ES1 gl, ActivityChangeType eventType)
	{
		if(eventType == ActivityChangeType.PAUSE_RESUME)
			return;
		
		if(transitionSteps > 0 && transition != TransitionKind.NONE) {
			currentTransition = new Transition(transition, transitionSteps, makeTransitionImage());
			autoDestroy(currentTransition);
		}
		transition = TransitionKind.NONE;
		transitionSteps = 0;
		
		CURRENT = this;
		EN.setTargetFPS(info.fps);
		if(info.caption != null)
			EN.setTitle(info.caption);
		if(!initd)
		{
			initd = true;
			if(info.persistent)
				persistentRooms.put(info.name, this);
			
			tiles.init(tileInfo);
	
			//Persistent instances need to be pre-added,
			//but don't call their create events again,
			//and don't recreate persistent instances that
			//came from an earlier instance of this room
			boolean[] alreadyExists = new boolean[instanceInfo.length];  
			for (Entity e : persistentInstances)
			{
				if(info.name.equals(e.persistenceSourceRoomName))
					alreadyExists[e.persistenceSourceInstanceIndex] = true;
				add(e);
			}
			iterator = instances.size();
			for(int i = 0; i<instanceInfo.length; i++)
			{
				if(!alreadyExists[i])
					newInstance(i);
			}
			iteratorMax = instances.size();
			while(iterator < iteratorMax)
			{
				Entity instance = instances.get(iterator);
				if(!instance.isDisposed())
					instance.createEvent();
				iterator++;
			}
	
			if(eventType == ActivityChangeType.GAME_START || eventType == ActivityChangeType.GAME_RESTART)
			{
				iterator = 0;
				iteratorMax = instances.size();
				while(iterator < iteratorMax)
				{
					Entity instance = instances.get(iterator);
					if(instance.isActive())
						instance.gameStart();
					iterator++;
				}
			}
	
			if(hasCreationCode && !EN.gameConfig.creationCodes.runCreationCode(info.name))
				System.err.println("Warning: Creation code not handled for: \""+info.name+"\"");
		}
	
		iterator = 0;
		iteratorMax = instances.size();
		while(iterator < iteratorMax)
		{
			Entity instance = instances.get(iterator);
			if(instance.isActive())
				instance.roomStart();
			iterator++;
		}

		EN.resetFrameTime();
		if(currentTransition != null)
		{
			render();
			currentTransition.init(makeTransitionImage());
		}
		stepViews();
	}
	
	protected Tex2D makeTransitionImage()
	{
		IImage prevImg = EN.fbm.screenshot();
		return new Tex2D(EN.gl, prevImg, false, TexSampler.CLAMP);
	}

	public void setRoomSpeed(int fps)
	{
		EN.setTargetFPS(fps);
		info.fps = fps;
	}

	public void setCaption(String caption)
	{
		info.caption = caption;
		EN.setTitle(caption);
	}

	@Override
	public void run(GL2ES1 gl)
	{
		if(currentTransition != null)
		{
			EN.clear();
//			float[] view = new float[4];
//			Letterbox.getViewport(view, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight(), currentTransition.next.dataWidth, currentTransition.next.dataHeight);
//			EN.gl.glViewport(round(view[0]), round(view[1]), round(view[2]), round(view[3]));
			EN.gl.glViewport(0, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight());
			if(currentTransition.render())
			{
				autoDestructibles.remove(currentTransition);
				currentTransition.dispose();
				currentTransition = null;
			}
		}
		else
		{
			if(EN.isFocused() || !EN.gameConfig.freezeOnFocusLoss)
			{
				step();
				render();
			}
			EN.clear();
			EN.fbm.blit(EN.getSurfaceWidth(), EN.getSurfaceHeight());
		}
	}

	protected Entity newInstance(int index)
	{
		InstanceInfo inf = instanceInfo[index];
		try
		{
			Class<? extends Entity> type = object(inf.object);
			Constructor<? extends Entity> constructor = type.getConstructor(double.class, double.class);
			Entity e = constructor.newInstance((double)inf.x, (double)inf.y);
			e.image_xscale = inf.xscale;
			e.image_yscale = inf.yscale;
			e.image_angle = inf.angle;
			e.image_blend = inf.color;
			e.image_alpha = ((inf.color>>>24)&0xFF)/255.0;
			if(inf.creationCode != null)
				e.initID(inf.id, inf.creationCode);
			e.persistenceSourceRoomName = info.name;
			e.persistenceSourceInstanceIndex = index;
			return e;
		}
		catch(Exception e)
		{
			throw new RuntimeException("Failed to create instance of "+inf.object, e);
		}
	}

	@Override
	public void end(GL2ES1 gl, ActivityChangeType eventType)
	{
		if(eventType == ActivityChangeType.PAUSE_RESUME)
			return;
		
		iterator = 0;
		iteratorMax = instances.size();
		while(iterator < iteratorMax)
		{
			Entity instance = instances.get(iterator);
			if(instance.isActive())
				instance.roomEnd();
			iterator++;
		}

		if(eventType == ActivityChangeType.GAME_END || eventType == ActivityChangeType.GAME_RESTART)
		{
			iterator = 0;
			iteratorMax = instances.size();
			while(iterator < iteratorMax)
			{
				Entity instance = instances.get(iterator);
				if(instance.isActive())
					instance.gameEnd();
				iterator++;
			}
		}
		if(!info.persistent)
		{
			dispose();
			iterator = 0;
			iteratorMax = instances.size();
			while(iterator < iteratorMax)
			{
				Entity instance = instances.get(iterator);
				if(!instance.persistent)
					instance.dispose();
				iterator++;
			}
		}
	}
	
	@Override
	public void dispose()
	{
		super.dispose();
		for(Entity e : persistentInstances)
			e.removeNotify(this);
	}
	
	public void add(Entity entity)
	{
		assert !isDisposed() : "Illegal instance creation after room end: "+entity;
		active.add(entity);
		instances.add(entity);
		entity.addNotify(this);
		if(entity.persistent)
			persistentInstances.add(entity);
	}

	public void disposeNotify(Entity e)
	{
		if(render)
			throw new RuntimeException("Illegal destruction during render: "+e);
		if(e.persistent)
			persistentInstances.remove(e);
		int index = instances.indexOf(e);
		if(index != -1)
		{
			if(e.isActive())
				active.remove(e);
			else
				inactive.remove(e);
			instances.remove(index);
			if(index < iteratorMax)
				iteratorMax--;
			if(index <= iterator)
				iterator--;
		}
	}
	
	protected static final ArrayList<KeyListener> klCache = new ArrayList<>(16);
	protected static final ArrayList<MouseListener> mlCache = new ArrayList<>(16);
	protected static final ArrayList<MouseIntersector> miCache = new ArrayList<>(16);
	protected static final ArrayList<Collider> clCache = new ArrayList<>(64);

	public void step()
	{
		render = false;
		
		iterator = 0;
		iteratorMax = instances.size();
		while(iterator < iteratorMax)
		{
			Entity instance = instances.get(iterator);
			if(instance.isActive())
			{
				//Unknown if this should be interleaved
				instance.setPrevious();
				instance.beginStep();
				instance.doTiming();
			}
			iterator++;
		}
		
		InputState in = EN.in;
		
		try {
			instances(klCache, KeyListener.class);
			for(int i = 0; i<in.getKeyEvents().size(); i++) {
				for(int j = 0; j<klCache.size(); j++)
				{
					KeyListener e = klCache.get(j);
					if(!e.isDisposed() && i < in.getKeyEvents().size())
						e.keyEvent(in.getKeyEvents().types.get(i), in.getKeyEvents().ids.get(i));
				}
			}
			instances(mlCache, MouseListener.class);
			for(int i = 0; i<in.getMouseEvents().size(); i++) {
				for(int j = 0; j<mlCache.size(); j++)
				{
					MouseListener e = mlCache.get(j);
					if(!e.isDisposed() && i < in.getMouseEvents().size())
						e.mouseEvent(in.getMouseEvents().types.get(i), in.getMouseEvents().ids.get(i));
				}
			}
			instances(miCache, MouseIntersector.class);
			for(int j = 0; j<miCache.size(); j++)
			{
				MouseIntersector e = miCache.get(j);
				if(!e.isDisposed())
					e.checkMouseHover();
			}

			//Profiler<Class<?>> prof = new Profiler<>("step", Profiler.classToString);
			iterator = 0;
			iteratorMax = instances.size();
			while(iterator < iteratorMax)
			{
				Entity instance = instances.get(iterator);
				if(instance.isActive())
					instance.step();
				iterator++;
			}
			//prof.print();
			
			iterator = 0;
			iteratorMax = instances.size();
			while(iterator < iteratorMax)
			{
				Entity instance = instances.get(iterator);
				if(instance.isActive())
					instance.doPhysics();
				iterator++;
			}
			
			instances(clCache, Collider.class);
			for(int j = 0; j<clCache.size(); j++)
			{
				Collider e = clCache.get(j);
				if(e.isDisposed())
					continue;
				e.handleCollisions();
				e.endCollisions();
			}
		} finally {
			klCache.clear();
			mlCache.clear();
			miCache.clear();
			clCache.clear();
		}
		
		iterator = 0;
		iteratorMax = instances.size();
		while(iterator < iteratorMax)
		{
			Entity instance = instances.get(iterator);
			if(instance.isActive())
			{
				instance.endStep();
				instance.doAnimation();
			}
			iterator++;
		}
		
		for(int i = 0; i<backgrounds.length; i++)
		{
			backgrounds[i].x += backgrounds[i].hspeed;
			backgrounds[i].y += backgrounds[i].vspeed;
		}
		
		stepViews();
	}
	
	public void stepViews()
	{
		if(info.viewsEnabled)
		{
			for(int i = 0; i<views.length; i++)
			{
				ViewData v = views[i];
				Entity obj = v.getObject();
				if(v.visible && obj != null)
				{
					ObjectTransform tf = obj.validateTransform();
					double ddr = v.xview+v.wview-v.hborder - tf.x1;
					if(ddr < 0)
					{
						if(v.hspeed >= 0)
							ddr = Math.min(ddr, v.hspeed);
						v.xview -= ddr;
					}
					double ddl = v.xview+v.hborder - tf.x0; 
					if(ddl > 0)
					{
						if(v.hspeed >= 0)
							ddl = Math.max(ddl, -v.hspeed);
						v.xview -= ddl;
					}
					
					double ddb = v.yview+v.hview-v.vborder - tf.y1;
					if(ddb < 0)
					{
						if(v.vspeed >= 0)
							ddb = Math.min(ddb, v.vspeed);
						v.yview -= ddb;
					}
					double ddu = v.yview+v.vborder - tf.y0; 
					if(ddu > 0)
					{
						if(v.vspeed >= 0)
							ddu = Math.max(ddu, -v.vspeed);
						v.yview -= ddu;
					}
					
					v.xview = Clamp.clampC(v.xview, 0, info.width-v.wview);
					v.yview = Clamp.clampC(v.yview, 0, info.height-v.hview);
				}
			}
		}
	}
	
	protected static int[] viewportSize = new int[2];
	
	public void getViewportSize(int[] output, int off)
	{
		int viewportWidth, viewportHeight;
		if(info.viewsEnabled) {
			viewportWidth = 0;
			viewportHeight = 0;
			for(int i = 0; i<views.length; i++) {
				if(views[i].visible) {
					viewportWidth = Math.max(viewportWidth, views[i].xport+views[i].wport);
					viewportHeight = Math.max(viewportHeight, views[i].yport+views[i].hport);
				}
			}
		} else {
			viewportWidth = info.width;
			viewportHeight = info.height;
		}
		output[off  ] = viewportWidth;
		output[off+1] = viewportHeight;
	}
	
	public boolean isRendering()
	{
		return render;
	}

	public void render()
	{
		render = true;
		getViewportSize(viewportSize, 0);
		
		EN.fbm.configure(EN.gl, viewportSize[0], viewportSize[1]);
		
		if(EN.fbm.beginRender(EN.getSurfaceWidth(), EN.getSurfaceHeight())) {
			EN.gl.glEnable(GL2ES1.GL_BLEND);
			EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
			BlendMode.NORMAL.set();
			sortedInstances.ensureCapacity(instances.size()+tiles.layers.size());
			for(int i = 0; i<instances.size(); i++)
			{
				Entity instance = instances.get(i);
				if(instance.isActive())
					sortedInstances.add(instance);
			}
			sortedInstances.addAll(tiles.layers);
			
			Collections.sort(sortedInstances, Renderable.DEPTH_COMPARATOR);
			
			if(info.viewsEnabled)
			{
				for(int i = views.length-1; i >= 0; i--)
				{
					ViewInfo v = views[i];
					if(v.visible)
						renderView(i, v.xport, v.yport, v.wport, v.hport, v.xview, v.yview, v.wview, v.hview);
				}
			}
			else
				renderView(-1, 0, 0, info.width, info.height, 0, 0, info.width, info.height);
			viewIndex = 0;
			sortedInstances.clear();
			EN.fbm.endRender();
		}
		render = false;
	}
	
	public int getViewIndex()
	{
		return viewIndex;
	}

	protected void renderView(int index, int xport, int yport, int wport, int hport, int xview, int yview, int wview, int hview) {
		viewIndex = index;
		EN.gl.glViewport(xport, yport, wport, hport);
		EN.gl.glMatrixMode(GL2ES1.GL_PROJECTION);
		EN.gl.glLoadIdentity();
		
		if(d3d_get_perspective())
		{
			EN.glu.gluPerspective(45, (float)wview/hview, 1, 65535);
			EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			EN.gl.glLoadIdentity();
			EN.glu.gluLookAt(xview+wview/2.0f, yview+hview/2.0f, wview, xview+wview/2.0f, yview+hview/2.0f, 0, 0, 1, 0);
		}
		else
		{
			EN.gl.glOrthof(xview, xview+wview, yview, yview+hview, -32768, 32767);
			EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			EN.gl.glLoadIdentity();
		}

		EN.gl.glScissor(xport, yport, wport, hport);
		EN.gl.glEnable(GL2ES1.GL_SCISSOR_TEST);

		drawBackgrounds();
		for(int i = 0; i<sortedInstances.size(); i++)
		{
			Renderable e = sortedInstances.get(i);
			if(e.visible)
				e.render();
		}
		drawForegrounds();
		EN.gl.glDisable(GL2ES1.GL_SCISSOR_TEST);
	}

	protected void drawBackgrounds()
	{
		if(d3d_get_perspective())
		{
			EN.gl.glClearDepth(1.0f);
			EN.gl.glClear(GL2ES1.GL_DEPTH_BUFFER_BIT);
		}

		if(info.drawBackgroundColor)
		{
			EN.gl.glClearColor(
					color_get_red(info.backgroundColor)/255f, 
					color_get_green(info.backgroundColor)/255f, 
					color_get_blue(info.backgroundColor)/255f, 
					1.0f
					);
			EN.gl.glClear(GL2ES1.GL_COLOR_BUFFER_BIT);
		}
		for(int i = 0; i<backgrounds.length; i++)
		{
			BackgroundData bg = backgrounds[i];
			if(bg.visible && bg.background != null && !bg.foreground)
				drawBackground(bg);
		}
	}

	protected void drawForegrounds()
	{
		for(int i = 0; i<backgrounds.length; i++)
		{
			BackgroundData bg = backgrounds[i];
			if(bg.visible && bg.background != null && bg.foreground)
				drawBackground(bg);
		}
	}

	protected void drawBackground(BackgroundData bg)
	{
		if(bg.stretch)
			draw_background_stretched_ext(bg.getBackground(), 0, 0, room_width(), room_height(), c_white, bg.alpha);
		else
			draw_background_tiled_ext(bg.getBackground(), bg.x, bg.y, bg.tileX, bg.tileY, c_white, bg.alpha);
	}

	@Override
	public String toString()
	{
		return info.name;
	}

	public void deactivateAll(Entity exception)
	{
		int i = 0;
		while(i<instances.size())
		{
			Entity instance = instances.get(i);
			if(instance != exception)
				deactivate(instance);
			//else
				i++;
		}
	}
	public void deactivate(Class<?> type)
	{
		int i = 0;
		while(i<instances.size())
		{
			Entity instance = instances.get(i);
			if(type.isInstance(instance))
			{
				deactivate(instance);
			}
			//else
				i++;
		}
	}
	public void deactivate(Entity instance)
	{
		if(active.remove(instance))
		{
			instance.activateNotify(false);
			inactive.add(instance);
		}
	}

	public void activate(Class<?> type)
	{
		IOrderSet<?> src = inactive.getInstances(type);
		Object n;
		while((n = src.first())!=null)
			activate((Entity)n);
	}
	public void activate(Entity instance)
	{
		if(inactive.remove(instance))
		{
			instance.activateNotify(true);
			active.add(instance);
		}
	}

	public void destroyAt(double x, double y)
	{
		int i = 0;
		while(i<instances.size())
		{
			Entity e = instances.get(i);
			Sprite m = e.getMask();
			if(m != null)
			{
				ObjectTransform tf = e.validateTransform();
				int mi = Modulo.mod((int)e.image_index, m.number);
    			if(tf.x0 <= x && tf.x1 >= x && tf.y0 <= y && tf.x1 >= y && m.getBit(tf, mi, x, y))
    			{
    				e.destroy();
    				i--;
    			}
			}
			i++;
		}
	}
}
