package owg.lgm.room;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.Background;
import owg.lgm.gl.AttribArray;
import owg.lgm.gl.BufferUsage;
import owg.lgm.gl.DataType;
import owg.lgm.gl.DrawCommand;
import owg.lgm.gl.IndexBufferObject;
import owg.lgm.gl.Mesh;
import owg.lgm.gl.PrimitiveType;
import owg.lgm.gl.VertexAttribute;
import owg.lgm.gl.VertexBufferObjectF;
import owg.lgm.gl.DrawCommand.DrawElements;
import owg.lgm.room.RoomData.TileInfo;
import owg.lgm.room.TileLayer.Batch.RenderTile;
import owg.lgm.util.AutoDestructibleImpl;
import owg.util.asserter.Asserter;
import owg.util.primitivelists.IntList;

public class TileLayer extends Renderable
{
	protected static final short[] I_MV_BUF = new short[6];
	protected static final float[] V_MV_BUF = new float[16];
	protected static final int ALLOC_BLK = 16;
	
	public static class Batch extends AutoDestructibleImpl
	{
		public static class RenderTile
		{
			public int vertexIndex;
			public int ibOffset;
		}
		public Background background;
		public LinkedHashMap<Integer, RenderTile> tilesByID;
		public ArrayList<RenderTile> allTiles;
		public ArrayList<RenderTile> visibleTiles;
		public FloatBuffer verts;
		public VertexBufferObjectF vbo;
		public ShortBuffer inds;
		public IndexBufferObject ibo;
		public Mesh mesh;
		
		public boolean iReplace, vReplace;
		public int iOutdated0, iOutdated1;
		public int vOutdated0, vOutdated1;
		
		public Batch()
		{
			//
		}
		
		public void init(GL2ES1 gl, Background bg, Collection<TileInfo> tiles)
		{
			this.background = bg;
			
			int initialCapacity = tiles.size()+ALLOC_BLK;
			tilesByID = new LinkedHashMap<>(initialCapacity);
			visibleTiles = new ArrayList<>();
			allTiles = new ArrayList<>();
			
			verts = ByteBuffer.allocateDirect(4*16*initialCapacity).order(ByteOrder.nativeOrder()).asFloatBuffer();
			inds = ByteBuffer.allocateDirect(2*6*initialCapacity).order(ByteOrder.nativeOrder()).asShortBuffer();
			
			int vi = 0;
			int ii = 0;
			for(TileInfo t : tiles)
			{
				RenderTile rt = new RenderTile();
				rt.vertexIndex = vi;
				rt.ibOffset = ii;
				tilesByID.put(t.id, rt);
				visibleTiles.add(rt);
				allTiles.add(rt);
				
				putVerts(t, rt);
				putInds(vi);
				
				vi += 4;
				ii += 6;
			}
			
			verts.clear();
			inds.clear();
			
			assert vbo == null;
			assert ibo == null;
			vbo = new VertexBufferObjectF(gl, verts, BufferUsage.DYNAMIC_DRAW);
			ibo = new IndexBufferObject(gl, inds, BufferUsage.DYNAMIC_DRAW);
			
			vReplace = false;
			iReplace = false;
			
			vOutdated0 = Integer.MAX_VALUE;
			vOutdated1 = 0;
			iOutdated0 = Integer.MAX_VALUE;
			iOutdated1 = 0;
			
			mesh = new Mesh(DrawCommand.drawElements(ibo, PrimitiveType.TRIANGLES, visibleTiles.size()*6, DataType.UNSIGNED_SHORT, 0), 
					new AttribArray(vbo, VertexAttribute.POSITION, 2, DataType.FLOAT, 4*4, 0),
					new AttribArray(vbo, VertexAttribute.TEXTURE_COORDINATE_0, 2, DataType.FLOAT, 4*4, 4*2));
			mesh.autoDestroy(vbo);
			mesh.autoDestroy(ibo);
			autoDestroy(mesh);
		}
		
		protected void ensureVertexCapacity(int tileCount)
		{
			int position = tileCount*16;
			if(position >= verts.capacity())
			{
				FloatBuffer r = ByteBuffer.allocateDirect(4*16*(tileCount+ALLOC_BLK)).order(ByteOrder.nativeOrder()).asFloatBuffer();
				verts.rewind();
				verts.limit(allTiles.size()*16);
				r.put(verts);
				verts = r;
				vReplace = true;
			}
		}
		
		protected void ensureIndexCapacity(int tileCount)
		{
			int position = tileCount*6;
			if(position >= inds.capacity())
			{
				ShortBuffer r = ByteBuffer.allocateDirect(2*6*(tileCount+ALLOC_BLK)).order(ByteOrder.nativeOrder()).asShortBuffer();
				inds.rewind();
				inds.limit(visibleTiles.size()*6);
				r.put(inds);
				inds = r;
				iReplace = true;
			}
		}
		
		protected void putVerts(TileInfo t, RenderTile rt)
		{
			float x0 = t.x, y0 = t.y;
			float x1 = x0+t.backgroundW, y1 = y0+t.backgroundH;
			
			float tx0 = ((t.backgroundX+0.5f)) / background.tex.texWidth ;
			float ty0 = ((t.backgroundY+0.5f)) / background.tex.texHeight;

			float tx1 = ((t.backgroundX+t.backgroundW-0.5f)) / background.tex.texWidth ;
			float ty1 = ((t.backgroundY+t.backgroundH-0.5f)) / background.tex.texHeight;
			
			verts.put(x0).put(y0).put(tx0).put(ty0);
			verts.put(x1).put(y0).put(tx1).put(ty0);
			verts.put(x0).put(y1).put(tx0).put(ty1);
			verts.put(x1).put(y1).put(tx1).put(ty1);
		}
		
		protected void vOutdated(RenderTile rt)
		{
			vOutdated0 = Math.min(vOutdated0, (rt.vertexIndex  )*4);
			vOutdated1 = Math.max(vOutdated1, (rt.vertexIndex+4)*4);
		}
		
		protected void putInds(int vi)
		{
			inds.put((short)(vi  )).put((short)(vi+2)).put((short)(vi+1));
			inds.put((short)(vi+1)).put((short)(vi+2)).put((short)(vi+3));
		}
		
		protected void iOutdated(RenderTile rt)
		{
			iOutdated0 = Math.min(iOutdated0, rt.ibOffset);
			iOutdated1 = Math.max(iOutdated1, rt.ibOffset+6);	
		}
		
		public void add(TileInfo t)
		{
			RenderTile rt = new RenderTile();
			rt.vertexIndex = allTiles.size()*4;
			rt.ibOffset = visibleTiles.size()*6;
			Asserter.check(tilesByID.put(t.id, rt) == null);
			allTiles.add(rt);
			visibleTiles.add(rt);

			ensureVertexCapacity(allTiles.size());
			verts.position(rt.vertexIndex*4);
			putVerts(t, rt);
			vOutdated(rt); 
			
			ensureIndexCapacity(visibleTiles.size());
			inds.position(rt.ibOffset);
			putInds(rt.vertexIndex);
			iOutdated(rt);
		}

		protected void hide(RenderTile tile)
		{
			//Move last tile to here
			RenderTile lastTile = visibleTiles.remove(visibleTiles.size()-1);
			if(lastTile == tile) {
				//Nothing to do
			} else {
				lastTile.ibOffset = tile.ibOffset;
				visibleTiles.set(tile.ibOffset / 6, lastTile);
				
				inds.position(lastTile.ibOffset);
				inds.get(I_MV_BUF);
				
				lastTile.ibOffset = tile.ibOffset;
				inds.position(lastTile.ibOffset);
				inds.put(I_MV_BUF);		
				iOutdated(lastTile);

				//Note: Old inds for last tile are out of bounds and do not immediately need update
			}
			tile.ibOffset = -1;
		}

		public boolean setVisible(int id, boolean visible)
		{
			RenderTile tile = tilesByID.get(id);
			if(tile != null)
			{
				if(visible)
				{
					if(tile.ibOffset == -1)
					{
						tile.ibOffset = visibleTiles.size()*6;
						visibleTiles.add(tile);
						
						ensureIndexCapacity(visibleTiles.size());
						inds.position(tile.ibOffset);
						putInds(tile.vertexIndex);
						iOutdated(tile);
					}
				}
				else if(tile.ibOffset != -1)
					hide(tile);
				return true;
			}
			return false;
		}
		public boolean remove(int id)
		{
			RenderTile tile = tilesByID.remove(id);
			if(tile != null)
			{
				if(tile.ibOffset != -1)
					hide(tile);

				//Move last tile to here
				RenderTile lastTile = allTiles.remove(allTiles.size()-1);
				if(lastTile == tile) {
					//Nothing to do
				} else {
					lastTile.vertexIndex = tile.vertexIndex;
					allTiles.set(tile.vertexIndex / 4, lastTile);
					
					verts.position(lastTile.vertexIndex*4);
					verts.get(V_MV_BUF);
					
					lastTile.vertexIndex = tile.vertexIndex;
					verts.position(lastTile.vertexIndex*4);
					verts.put(V_MV_BUF);		
					
					vOutdated(lastTile);

					//Also update inds for lastTile if it is visible
					if(lastTile.ibOffset != -1)
					{
						inds.position(lastTile.ibOffset);
						putInds(lastTile.vertexIndex);
						iOutdated(lastTile);
					}
				}
				return true;
			}
			return false;
		}

		public void getTiles(IntList output)
		{
			output.ensureCapacity(output.length+tilesByID.size());
			for(Integer id : tilesByID.keySet())
				output.add(id);
		}
		
		public void update()
		{
			if(vReplace)
			{
				verts.rewind();
				vbo.setData(verts);
				vReplace = false;
				vOutdated0 = Integer.MAX_VALUE;
				vOutdated1 = 0;
			}
			else if(vOutdated1 > vOutdated0)
			{
				vbo.update(vOutdated0, vOutdated1-vOutdated0);
				vOutdated0 = Integer.MAX_VALUE;
				vOutdated1 = 0;
			}
			if(iReplace)
			{
				inds.rewind();
				ibo.setData(inds);
				iReplace = false;
				iOutdated0 = Integer.MAX_VALUE;
				iOutdated1 = 0;
			}
			else if(iOutdated1 > iOutdated0)
			{
				ibo.update(iOutdated0, iOutdated1-iOutdated0);
				iOutdated0 = Integer.MAX_VALUE;
				iOutdated1 = 0;
			}
			
			((DrawElements)mesh.drawCommand).count = visibleTiles.size()*6;
		}
	}
	protected LinkedHashMap<Background, Batch> batches = new LinkedHashMap<>();

	public void init(GL2ES1 gl, int depth, Collection<TileInfo> tiles) 
	{
		this.depth = depth;
		this.visible = true;
		
		LinkedHashMap<String, ArrayList<TileInfo>> parts = new LinkedHashMap<>();
		for(TileInfo t : tiles)
		{
			String bg = t.backgroundName;
			ArrayList<TileInfo> list = parts.get(bg);
			if(list == null)
			{
				list = new ArrayList<>();
				parts.put(bg, list);
			}
			list.add(t);
		}
		
		for(Entry<String, ArrayList<TileInfo>> e : parts.entrySet())
		{
			Background bg = Background.get(e.getKey());
			Batch bat = new Batch();
			autoDestroy(bat);
			bat.init(gl, bg, e.getValue());
			batches.put(bg, bat);
		}
	}

	public void add(GL2ES1 gl, TileInfo tile)
	{
		Background bg = Background.get(tile.backgroundName);
		Batch bat = batches.get(bg);
		if(bat == null)
		{
			bat = new Batch();
			bat.init(gl, bg, Collections.emptyList());
			autoDestroy(bat);
			batches.put(bg, bat);
		}
		bat.add(tile);
	}

	public boolean remove(int id)
	{
		Iterator<Batch> it = batches.values().iterator();
		while(it.hasNext())
		{
			Batch b = it.next();
			if(b.remove(id))
			{
				if(b.allTiles.isEmpty())
				{
					b.dispose();
					it.remove();
				}
				return true;
			}
		}
		return false;
	}

	public boolean setVisible(int id, boolean visible)
	{
		for(Batch b : batches.values())
		{
			if(b.setVisible(id, visible))
				return true;
		}
		return false;
	}

	public void getTiles(IntList output)
	{
		for(Batch b : batches.values())
			b.getTiles(output);
	}
	
	public boolean getTileInfo(TileInfo output, int id)
	{
		for(Batch b : batches.values())
		{
			RenderTile rt = b.tilesByID.get(id);
			if(rt != null)
			{
				FloatBuffer buf = b.vbo.getData();
				int base = rt.vertexIndex*4;
				output.x = Math.round(buf.get(base  ));
				output.y = Math.round(buf.get(base+1));
				output.backgroundName = b.background.name;
				output.backgroundX = Math.round((buf.get(base+2) * b.background.tex.texWidth ) - 0.5f);
				output.backgroundY = Math.round((buf.get(base+3) * b.background.tex.texHeight) - 0.5f);
				output.backgroundW = Math.round((buf.get(base+14)* b.background.tex.texWidth ) + 0.5f - output.backgroundX);
				output.backgroundH = Math.round((buf.get(base+15)* b.background.tex.texHeight) + 0.5f - output.backgroundY);
				output.depth = depth;
				output.id = id;
				output.locked = false;
				return true;
			}
		}
		return false;
	}
	
	public boolean isEmpty()
	{
		return batches.isEmpty();
	}

	@Override
	public void render()
	{
		for(Entry<Background, Batch> e : batches.entrySet())
		{
			Background bg = e.getKey();
			Batch bat = e.getValue();
			bat.update();
			if(!bat.visibleTiles.isEmpty())
			{
				GL2ES1 gl = bat.vbo.getGL();
				gl.glColor4f(1, 1, 1, 1);
				bg.tex.begin();
				bat.mesh.render(gl);
				bg.tex.end();
			}
		}
	}
}
