package owg.lgm.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.NoSuchElementException;

import owg.util.primitivelists.IntList;

/**
 * A list where each element retains its index until it is removed, even if
 * previous elements are removed first. This list cannot hold null elements!
 */
public class FixedIndexList<E>
{
	/**
	 * The default initial capacity for the backing array, when using the
	 * no-args constructor.
	 */
	public static final int DEFAULT_INITIAL_CAPACITY = 16;

	private int capacity;
	private int size;
	private IntList deletedIndices = new IntList(64);
	private Object[] allInstances;

	/**
	 * Creates a new {@link FixedIndexList} with an initial capacity of
	 * {@value #DEFAULT_INITIAL_CAPACITY}.
	 */
	public FixedIndexList() {
		this(DEFAULT_INITIAL_CAPACITY);
	}

	/**
	 * Creates a new {@link FixedIndexList} with the given initial capacity.
	 * 
	 * @param initialCapacity
	 *            The initial size of the backing array.
	 */
	public FixedIndexList(int initialCapacity) {
		capacity = (int) initialCapacity;
		allInstances = new Object[capacity];
		size = 0;
	}

	/**
	 * Creates a new {@link FixedIndexList} with the given collection of
	 * initial elements. The elements are added in the order given by
	 * {@link Collection#toArray()}, with ids starting from 0.
	 */
	public FixedIndexList(Collection<? extends E> initialElements) {
		capacity = (int) initialElements.size();
		allInstances = initialElements.toArray();
		for(int i = 0; i<initialElements.size(); i++)
			assert allInstances[i] != null;
		size = (int) initialElements.size();
	}

	/**
	 * Assigns the given element to the given index in the list, resizing the
	 * list if necessary. The indicated index must be empty!
	 * 
	 * @param index
	 *            The index at which to put the element.
	 * @param e
	 *            The element to put into the list.
	 * @throws IllegalStateException
	 *             If the index is already occupied.
	 */
	public void set(int index, E e)
	{
		if (index >= size)
		{
			if (index >= capacity)
				ensureCapacity((int) ((index + 1) * 2));
			size = (int) (index + 1);
		}
		else
		{
			if (allInstances[index] != null)
				throw new IllegalStateException("The list already has an element (" + allInstances[index] + ") at index: " + index);
			// Ensure this will not be overwritten by a subsequent add
			for(int i = 0; i<deletedIndices.length; i++)
			{
				if(deletedIndices.data[i] == index)
				{
					deletedIndices.data[i] = deletedIndices.data[deletedIndices.length-1];
					deletedIndices.length--;
					break;
				}
			}
		}

		allInstances[index] = e;
	}

	/**
	 * Adds the given element to the list and returns the index where it was
	 * added.
	 * 
	 * @param e
	 *            The element to add. <b>Must be non-null.</b>
	 * @return The index of the added element.
	 */
	public int add(E e)
	{
		assert e != null;
		int index;
		if (deletedIndices.length == 0)
		{
			if (size >= capacity)
				ensureCapacity((int) (1+size * 2));
			assert size < capacity;
			index = size;
			size++;
		}
		else
		{
			index = deletedIndices.data[deletedIndices.length-1];
			deletedIndices.length--;
		}
		allInstances[index] = e;
		return index;
	}

	/**
	 * Removes the given element, if it exists.
	 * 
	 * @param e
	 *            The element to remove.
	 * @return The index of the removed element, -1 if it did not exist.
	 */
	public int remove(E e)
	{
		for (int i = 0; i < size; i++)
		{
			if (allInstances[i] == e)
			{
				removeIndex(i);
				return i;
			}
		}
		return -1;
	}

	/**
	 * Returns whether there is an element at the given index. If the index is
	 * out of bounds, this method returns false.
	 * 
	 * @param index
	 *            The index in which to look for an element.
	 * @return True if there is an element at the given index, false otherwise.
	 */
	public boolean containsIndex(int index)
	{
		return index >= 0 && index < size && allInstances[index] != null;
	}

	/**
	 * Gets the element at an index. This method is fail-fast, call
	 * {@link #containsIndex(int)} if the existence of the index is unknown.
	 * 
	 * @param index
	 *            The index of the instance to return
	 * @return The instance at the given index.
	 * @Throws {@link NoSuchElementException},
	 *         {@link ArrayIndexOutOfBoundsException}
	 */
	public E get(int index)
	{
		Object obj = allInstances[index];
		if (obj == null)
			throw new NoSuchElementException("No element at index: " + index);
		@SuppressWarnings("unchecked")
		E tmp = (E) allInstances[index];
		return tmp;
	}

	/**
	 * Removes the element at the given index. This method is fail-fast, call
	 * {@link #containsIndex(int)} if the existence of the index is unknown.
	 * 
	 * @param index
	 *            The index of the element to remove.
	 * @throws NoSuchElementException
	 */
	public E removeIndex(int index)
	{
		Object obj = allInstances[index];
		if (obj == null)
			throw new NoSuchElementException("No element at index: " + index);
		@SuppressWarnings("unchecked")
		E tmp = (E) allInstances[index];
		deletedIndices.add(index);
		allInstances[index] = null;
		return tmp;
	}

	/**
	 * Returns whether the list contains the given element.
	 * 
	 * @param e
	 *            The element to search for.
	 * @return Whether the given element is in the list.
	 */
	public boolean contains(E e)
	{
		for (int i = 0; i < size; i++)
		{
			if (allInstances[i] == e)
				return true;
		}
		return false;
	}

	/**
	 * Ensures that the current capacity of the backing array is at least the
	 * given minimum, resizing the backing array if necessary.
	 */
	protected void ensureCapacity(int minimum)
	{
		if (capacity < minimum)
		{
			allInstances = Arrays.copyOf(allInstances, minimum);
			capacity = minimum;
		}
	}

	/**
	 * Returns the number of used slots in the list. Note that this includes
	 * slots that have been occupied, but are now deleted!
	 */
	public int size() {
		return size;
	}
	
	/**Clear the list, freeing up all slots.*/
	public void clear()
	{
		Arrays.fill(allInstances, null);
		deletedIndices.length = 0;
		size = 0;
	}

	@Override
	public String toString() {
		return Arrays.toString(allInstances);
	}
}