package com.oddwarg.gervill;

import java.util.Map;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Soundbank;
import javax.sound.midi.Track;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;

//This file is public domain

/**Enhanced version of {@link SoftSynthesizer}, adds:<ul>
 * <li>More efficient and less confusing API for using a custom soundbank: {@link #EnhancedSoftSynthesizer(Soundbank)}</li>
 * <li>Loading a midi sequence with precise timestamps: {@link #send(Sequence, long)}</li>
 * <li>Stopping the midi sequence: {@link #stop()} (incomplete)</li>
 * </ul>*/
public class EnhancedSoftSynthesizer extends SoftSynthesizer
{
	protected class EnhancedReceiver extends SoftReceiver
	{
		public EnhancedReceiver(SoftSynthesizer synth)
		{
			super(synth);
		}
		
		public void clear()
		{
			midimessages.clear();
		}
	}

	protected final Soundbank defBank;
	protected EnhancedReceiver clearReceiver = null;

	public EnhancedSoftSynthesizer(Soundbank bank)
	{
		this.defBank = bank;
	}
	
	@Override
	public Soundbank getDefaultSoundbank()
	{
		if(defBank == null)
			return super.getDefaultSoundbank();
		else
			return defBank;
	}
	
	@Override
	public AudioInputStream openStream(AudioFormat targetFormat, Map<String, Object> info) throws MidiUnavailableException
	{
		AudioInputStream r = super.openStream(targetFormat, info);
		if(clearReceiver == null)
			clearReceiver = new EnhancedReceiver(this);
		return r;
	}
	
	@Override
	public void close()
	{
		super.close();
		if(clearReceiver != null)
		{
			clearReceiver.close();
			clearReceiver = null;
		}
	}
	
	/**
	 * Send entry MIDI Sequence into Receiver using timestamps.
	 * Based on code from JFugue - API for Music Programming<br>
	 * Copyright (C) 2003-2008  Karl Helgason and David Koelle<br>
	 */
	public long send(Sequence seq, long curtime) {
		Receiver recv = clearReceiver;
		
		float divtype = seq.getDivisionType();
		assert (seq.getDivisionType() == Sequence.PPQ);
		Track[] tracks = seq.getTracks();
		int[] trackspos = new int[tracks.length];
		int mpq = 500000;
		int seqres = seq.getResolution();
		long lasttick = 0;
		while (true) {
			MidiEvent selevent = null;
			int seltrack = -1;
			for (int i = 0; i < tracks.length; i++) {
				int trackpos = trackspos[i];
				Track track = tracks[i];
				if (trackpos < track.size()) {
					MidiEvent event = track.get(trackpos);
					if (selevent == null
							|| event.getTick() < selevent.getTick()) {
						selevent = event;
						seltrack = i;
					}
				}
			}

			if (seltrack == -1)
				break;
			trackspos[seltrack]++;
			long tick = selevent.getTick();
			if (divtype == Sequence.PPQ)
				curtime += ((tick - lasttick) * mpq) / seqres;
			else
				curtime = (long) ((tick * 1000000.0 * divtype) / seqres);
			lasttick = tick;
			MidiMessage msg = selevent.getMessage();
			if (msg instanceof MetaMessage) {
				if (divtype == Sequence.PPQ) {
					if (((MetaMessage) msg).getType() == 0x51) {
						byte[] data = ((MetaMessage) msg).getData();
						mpq = ((data[0] & 0xff) << 16)
								| ((data[1] & 0xff) << 8) | (data[2] & 0xff);
					}
				}
			} 
			else {
				if (recv != null)
					recv.send(msg, curtime);//Send starting from current position of synth(better than resetting the position using reflection)
			}
		}
		return curtime;
	}
	
	/**Stops the currently playing song*/
	public void stop()
	{
		//Clear the list of unprocessed midi messages
		if(clearReceiver != null)
			clearReceiver.clear();
		
		//Stop and reset midi state
		try {
			//Softsynthesizer does not respond to MIDI reset message 0xFF, which is very sad. 
			//We have to perform a manual reset.
			ShortMessage message = new ShortMessage();
			for(int ch = 0; ch < 16; ch++)
			{
				//All notes off
				message.setMessage(ShortMessage.CONTROL_CHANGE|ch, 123, 0);
				clearReceiver.send(message, -1);
				//All controllers off
				message.setMessage(ShortMessage.CONTROL_CHANGE|ch, 121, 0);
				clearReceiver.send(message, -1);
				//Need to reset the volume (default 100) separately:
				message.setMessage(ShortMessage.CONTROL_CHANGE|ch, 0x07, 100);
				clearReceiver.send(message, -1);
				
				//Notice:
				//This fails to reset pitch bend and many channel properties
				//For a complete solution the help of a midi expert is required 
			}
		} catch(InvalidMidiDataException e) {
			throw new RuntimeException(e); //dead code
		}
	}
}
